# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.device import Device

from pyaoscx.pyaoscx_module import PyaoscxModule


class QueueProfile(PyaoscxModule):
    """
    Provide configuration management for Queue profiles on AOS-CX devices.
    """

    collection_uri = "system/q_profiles"
    object_uri = collection_uri + "/{name}"
    resource_uri_name = "name"
    indices = ["name"]

    def __init__(self, session, name, **kwargs):
        self.session = session
        self.__name = name

        # List used to determine attributes related to the
        # Queue profile configuration
        self.config_attrs = []
        self.materialized = False

        # Attribute dictionary used to manage the original data
        # obtained from the GET request
        self._original_attributes = {}
        utils.set_creation_attrs(self, **kwargs)

        # Used to know if the object was changed since the last
        # request
        self.__modified = False

        # Build the URI that identifies the current Queue profile
        self.path = self.object_uri.format(name=self.name)
        self.base_uri = self.collection_uri

    @property
    def name(self):
        return self.__name

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Queue profile and fill
            the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information
            to return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Update the original attributes
        self._original_attributes = data

        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all Queue profiles and create a
            dictionary containing each of them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing Queue profile names as keys and a Queue
            profile object as value.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.collection_uri)
        except Exception as exc:
            raise ResponseError("GET", exc)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        profile_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            profile_name, profile = cls.from_uri(session, uri)
            profile_dict[profile_name] = profile

        return profile_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Queue Profile.
            Checks whether the Queue Profile exists in the switch and calls
            self.update() or self.create() accordingly.
        :return modified: True if the object was modified.
        """
        if self.materialized:
            return self.update()
        else:
            return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Queue Profile.
        :return modified: True if the object was modified and a PUT request was
            made.
        """
        data = utils.get_attrs(self, self.config_attrs)
        # Manually remove the name
        if "name" in data:
            del data["name"]
        self.__modified = self._put_data(data)
        logging.info("SUCCESS: Updating %s", self)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Queue profile in the switch.
        :return modified: True if the object was modified.
        """
        data = utils.get_attrs(self, self.config_attrs)
        # Manually add the name
        data["name"] = self.name
        self.__modified = self._post_data(data)
        logging.info("SUCCESS: Adding %s", self)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform a DELETE call to remove a Queue Profile from the switch.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        logging.info("SUCCESS: Deleting %s", self)
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Queue profile object given an URI.
        :param cls: Object's class.
        :param session: Pyaoscx.Session objec used to represent a logical
            connection to the device.
        :param uri: a string with the URI
        :return id, object: tuple with the name and the Profile.
        """
        # Obtain the ID from URI
        name = uri.split("/")[-1]
        return name, cls(session, name)

    @classmethod
    def get_facts(cls, session):
        """
        Retrieve the information of all Queue profiles.
        :param cls: Class reference.
        :param session: Pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing the name as key and the materialized
            object as value.
        """
        logging.info("Retrieving Queue Profiles facts")

        depth = session.api.default_facts_depth

        try:
            response = session.request(
                "GET", cls.collection_uri, params={"depth": depth}
            )
        except Exception as exc:
            raise ResponseError("GET", exc)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        return json.loads(response.text)

    def __str__(self):
        return "Queue Profile {0}".format(self.name)

    @property
    def modified(self):
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        return self.modified

    @classmethod
    def set_global_queue_profile(cls, session, profile):
        """
        Sets the global queue profile for the switch.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param profile: Name of a queue profile to set as the global queue
            profile.
        """
        logging.info("Setting global queue profile to: %s", profile)
        device = Device(session)
        if not device.materialized:
            device.get()
        setattr(device, "q_profile_default", profile)
        return device.apply()



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class QosCos(PyaoscxModule):
    """
    Provide configuration management for QoS COS trust mode on AOS-CX devices.
    """

    base_uri = "system/qos_cos_map_entries"
    resource_uri_name = "qos_cos_map_entries"

    indices = ["code_point"]

    def __init__(self, session, code_point, **kwargs):
        """
        Initialize a QoS COS trust mode object.
        :param session: pyaoscx.Session object used to represent logical
            connection to the device.
        :param code_point: Integer to identify a QoS COS trust mode object.
        """
        self.session = session
        self.__code_point = code_point

        # List used to determine attributes related to the QoS COS
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}

        # Attribute used to know if object was changed recently
        self.__modified = False
        # Build path
        self.path = "{0}/{1}".format(self.base_uri, self.code_point)

    @property
    def code_point(self):
        """
        Method used to retrieve object's code point.
        :return: returns the code point of this QoS COS trust mode object.
        """
        # This uses the @property decorator to make self.code_point read-only
        return self.__code_point

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a QoS COS trust mode table
            entry and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        if selector in self.session.api.configurable_selectors:
            utils.set_config_attrs(self, data, "config_attrs")

        # Set original attributes
        self._original_attributes = data

        self.__color = data["color"]
        self.__description = data["description"]
        self.__local_priority = data["local_priority"]

        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system QoS COS trust mode
            configurations from of a switch.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing all system Schedule Profiles.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        qos_cos_dict = {}
        data = json.loads(response.text)
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            code_point, qos_cos = cls.from_uri(session, uri)
            qos_cos_dict[code_point] = qos_cos

        return qos_cos_dict

    @PyaoscxModule.materialized
    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to update an existing QoS COS trust mode table entry.
            Checks whether the QoS COS trust mode exists in the switch. Calls
            self.update if object is being updated.
        :return modified: Boolean, True if object was modified.
        """
        # Modify object
        self.__modified = self.update()
        return self.__modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to update an existing QoS COS trust mode object.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        qos_cos_data = utils.get_attrs(self, self.config_attrs)
        return self._put_data(qos_cos_data)

    @PyaoscxModule.connected
    def create(self):
        # TODO: Remove when abstractmethod decorator is removed from parent
        pass

    @PyaoscxModule.connected
    def delete(self):
        # TODO: Remove when abstractmethod decorator is removed from parent
        pass

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a QoS COS trust mode object given a response_data related to the
            existing QoS COS trust mode object.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                "3" : "/rest/v10.08/system/qos_cos_map_entries/3"
            }
        :return: QoS COS trust mode object.
        """
        code_points_arr = session.api.get_keys(
            response_data, cls.resource_uri_name
        )
        code_point = code_points_arr[0]

        return cls(session, code_point)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create an object given a QoS COS trust mode URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: s String with an URI.
        :return: returns identifier and object.
        """
        # Separate values from URI
        if cls.base_uri not in uri:
            raise ValueError(
                "ERROR: Invalid URI. String must be a valid QoS COS trust"
                "mode URI."
            )

        # Extract code point from URI
        code_point = uri.split("/")[-1]
        qos_cos = cls(session, code_point)

        # Return identifier and object
        return code_point, qos_cos

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain object's URI.
        :return: Object's URI.
        """
        # The self.path variable stores the URI
        return self.path

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method to check it object has been modified.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def __str__(self):
        return "QoS COS trust mode {0}".format(self.code_point)

    @property
    def color(self):
        """
        Getter method for the color property.
        """
        return self.__color

    @color.setter
    def color(self, color):
        """
        Updates the value of the color of this QoS COS instance.
        :param color: String to identify the color which may be used later in
            the pipeline in packet-drop decision points. Example: "green".
        """
        # Verify data type
        if not isinstance(color, str):
            raise ValueError("ERROR: Color value must be a string.")

        self.__color = color

    @property
    def description(self):
        """
        Getter method for the description property.
        """
        return self.__description

    @description.setter
    def description(self, description):
        """
        Updates the description of this QoS COS instance.
        :param description: String used for customer documentation.
        """
        # Verify data type
        if not isinstance(description, str):
            raise ValueError("ERROR: Description value must be a string.")

        self.__description = description

    @property
    def local_priority(self):
        """
        Getter method for the local_priority.
        """
        return self.__local_priority

    @local_priority.setter
    def local_priority(self, priority):
        """
        Updates the value of the local priority of this QoS COS instance.
        :param priority: Integer to represent an internal meta-data value that
            will be associated with the packet. This value will be used later
            to select the egress queue for the packet.
        """
        # Verify data type
        if not isinstance(priority, int):
            raise ValueError("ERROR: Priority must be an integer.")

        self.__local_priority = priority



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class BgpNeighbor(PyaoscxModule):
    """
    Provide configuration management for BGP Neighbor on AOS-CX devices.
    """

    indices = ["ip_or_ifname_or_group_name"]
    resource_uri_name = "bgp_neighbors"

    def __init__(
        self,
        session,
        ip_or_ifname_or_group_name,
        parent_bgp_router,
        uri=None,
        **kwargs
    ):

        self.session = session
        # Assign ID
        self.ip_or_ifname_or_group_name = ip_or_ifname_or_group_name
        # Assign parent BGP Router
        self.__set_bgp_router(parent_bgp_router)
        self._uri = uri
        # List used to determine attributes related to the BGP configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_bgp_router(self, parent_bgp_router):
        """
        Set parent BgpRouter object as an attribute for the BGP class
        :param parent_bgp_router a BgpRouter object
        """

        # Set parent BGP Router
        self.__parent_bgp_router = parent_bgp_router

        # Set URI
        self.base_uri = "{0}/{1}/bgp_neighbors".format(
            self.__parent_bgp_router.base_uri, self.__parent_bgp_router.asn
        )

        for bgp_ngh in self.__parent_bgp_router.bgp_neighbors:
            if (
                bgp_ngh.ip_or_ifname_or_group_name
                == self.ip_or_ifname_or_group_name
            ):
                # Make list element point to current object
                bgp_ngh = self
            else:
                # Add self to BGP Neighbors list in parent BGP Router
                self.__parent_bgp_router.bgp_neighbors.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a BGP Neighbor table
        entry and fill the object with the incoming attributes

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.ip_or_ifname_or_group_name)

        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the BGP Neighbor is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["ip_or_ifname_or_group_name"]
            )

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "ip_or_ifname_or_group_name" in self.__original_attributes:
            self.__original_attributes.pop("ip_or_ifname_or_group_name")

        # If the BGP Neighbor has a local_interface inside the switch
        if hasattr(self, "local_interface") and self.local_interface:
            local_interface_response = self.local_interface
            interface_cls = self.session.api.get_module(
                self.session, "Interface", ""
            )
            # Set port as a Interface Object
            self.local_interface = interface_cls.from_response(
                self.session, local_interface_response
            )
            self.local_interface.get()

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_bgp_router):
        """
        Perform a GET call to retrieve all system BGP Neighbors inside a BGP
        Router, and create a dictionary containing them
        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param parent_bgp_router: parent BgpRouter object where BGP Neighbor
            is stored
        :return: Dictionary containing BGP Neighbors IDs as keys and a BGP
            Neighbors objects as values
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/bgp_neighbors".format(
            parent_bgp_router.base_uri, parent_bgp_router.asn
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        bgp_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a BgpNeighbor object
            ip_or_ifname_or_group_name, bgp_neighbor = BgpNeighbor.from_uri(
                session, parent_bgp_router, uri
            )
            # Load all BGP Neighbor data from within the Switch
            bgp_neighbor.get()
            bgp_dict[ip_or_ifname_or_group_name] = bgp_neighbor

        return bgp_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing BGP Neighbor.
            Checks whether the BGP Neighbor exists in the switch. Calls
            self.update() if BGP Neighbor is being updated. Calls self.create()
            if a new BGP Neighbor is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_bgp_router.materialized:
            self.__parent_bgp_router.apply()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing BGP Neighbor.
        :return modified: True if Object was modified and a PUT request
            was made.
        """
        # Variable returned
        modified = False

        bgp_neighbor_data = utils.get_attrs(self, self.config_attrs)

        # Get ISL port uri
        if self.local_interface is not None:
            bgp_neighbor_data[
                "local_interface"
            ] = self.local_interface.get_info_format()

        uri = "{0}/{1}".format(self.base_uri, self.ip_or_ifname_or_group_name)

        # Compare dictionaries
        if bgp_neighbor_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            put_data = json.dumps(bgp_neighbor_data)

            try:
                response = self.session.request("PUT", uri, data=put_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)
            # Set new original attributes
            self.__original_attributes = bgp_neighbor_data
            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new BGP Neighbor table entry
        Only returns if an exception is not raise

        :return modified: Boolean, True if entry was created

        """
        bgp_neighbor_data = utils.get_attrs(self, self.config_attrs)
        bgp_neighbor_data[
            "ip_or_ifname_or_group_name"
        ] = self.ip_or_ifname_or_group_name

        if hasattr(self, "local_interface"):

            # If local interface is NOT a string
            if not isinstance(self.local_interface, str):
                if not self.local_interface.materialized:
                    raise VerificationError(
                        "Local Interface", "Object not materialized"
                    )

                # Get ISL port uri
                bgp_neighbor_data[
                    "local_interface"
                ] = self.local_interface.get_info_format()

        post_data = json.dumps(bgp_neighbor_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()

        # Object was modified, as it was created inside Device
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete  BGP Neighbor table entry.

        """

        uri = "{0}/{1}".format(self.base_uri, self.ip_or_ifname_or_group_name)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from BGP_Routers
        for neighbor in self.__parent_bgp_router.bgp_neighbors:
            if (
                neighbor.ip_or_ifname_or_group_name
                == self.ip_or_ifname_or_group_name
            ):
                self.__parent_bgp_router.bgp_neighbors.remove(neighbor)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_bgp_router, response_data):
        """
        Create a  BgpNeighbor object given a response_data related to the
        BGP Router ID object
        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param parent_bgp_router: parent BgpRouter object where BGP
            Neighbor is stored
        :param response_data: The response can be either a
            dictionary: {
                    id: "/rest/v10.04/system/vrfs/<vrf_name>/bgp_routers/asn
                        /bgp_neighbors/id"
                }
            or a
            string: "/rest/v10.04/system/vrfs/<vrf_name>/bgp_routers/asn/
                bgp_neighbors/id"
        :return: BgpNeighbor object
        """
        bgp_arr = session.api.get_keys(
            response_data, BgpNeighbor.resource_uri_name
        )
        bgp_neighbor_id = bgp_arr[0]
        return BgpNeighbor(session, bgp_neighbor_id, parent_bgp_router)

    @classmethod
    def from_uri(cls, session, parent_bgp_router, uri):
        """
        Create a BgpNeighbor object given a URI
        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param parent_bgp_router: parent BgpRouter object where BGP Neighbor
            is stored
        :param uri: a String with a URI

        :return index, bgp_obj: tuple containing both the BGP object and the
            BGP's ID
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)bgp_neighbors/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create BGP object
        bgp_obj = BgpNeighbor(session, index, parent_bgp_router, uri=uri)

        return index, bgp_obj

    def __str__(self):
        return "Bgp Neighbor ID {0}".format(self.ip_or_ifname_or_group_name)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific BGP Neighbor URI
        return: Object's URI
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.ip_or_ifname_or_group_name,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
        other objects
        return: Object format depending on the API Version
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute
        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.pyaoscx_module import PyaoscxModule


class OspfRouter(PyaoscxModule):
    """
    Provide configuration management for OSPF Routers on AOS-CX devices.
    """

    version = ""
    collection_uri = "system/vrfs/{name}/ospf{version}_routers"
    object_uri = collection_uri + "/{instance_tag}"

    indices = ["instance_tag"]
    resource_uri_name = "ospf_routers"

    # Use to manage references
    areas = ListDescriptor("areas")

    def __init__(self, session, instance_tag, parent_vrf, **kwargs):
        self.session = session
        # Assign ID
        self.__instance_tag = instance_tag
        # Assign parent Vrf object
        self._parent_vrf = parent_vrf
        # List used to determine attributes related to the OSPF configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary to manage original data obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        self.passive_interfaces = None
        if kwargs.get("passive_interfaces"):
            self.passive_interfaces = [
                self.session.api.get_module(self.session, "Interface", i)
                for i in kwargs["passive_interfaces"]
            ]
        # Use to manage Areas
        self.areas = []
        # Attribute used to know if object was changed recently
        self.__modified = False
        uri_indices = {
            "name": self._parent_vrf.name,
            "version": self.version,
            "instance_tag": self.__instance_tag,
        }
        self.base_uri = self.collection_uri.format(**uri_indices)
        self.path = self.object_uri.format(**uri_indices)
        self._parent_vrf.update_ospf_routers(self)

    @property
    def instance_tag(self):
        """
        Return this OSPF Router's instance_tag.
        """
        return self.__instance_tag

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a OSPF Router table entry and
            fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)
        # this is common for all PyaoscxModule derived classes
        self._get_and_copy_data(depth, selector, self.indices)
        # Delete unwanted data
        if "areas" in self._original_attributes:
            del self._original_attributes["areas"]
        # Set original attributes
        if "instance_tag" in self._original_attributes:
            del self._original_attributes["instance_tag"]
        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        # Set a list of passive_interfaces as an attribute
        if self.passive_interfaces:
            interfaces_list = []
            # Get all URI elements in the form of a list
            uri_list = self.session.api.get_uri_from_data(
                self.passive_interfaces
            )
            # gotta use deferred import to avoid cyclical import error
            from pyaoscx.interface import Interface

            for uri in uri_list:
                # Create an Interface object
                _, iface = Interface.from_uri(self.session, uri)
                # Add interface to list
                interfaces_list.append(iface)
            # Set list as Interfaces
            self.passive_interfaces = interfaces_list
        if self.areas == []:
            # Set Areas if any
            # Adds Area to parent OspfRouter
            from pyaoscx.ospf_area import OspfArea

            OspfArea.get_all(self.session, self)
        return True

    @classmethod
    def get_all(cls, session, parent_vrf):
        """
        Perform a GET call to retrieve all system OSPF Router settings for a
            given VRF, and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: Vrf object where OspfRouter object is stored.
        :return: Dictionary containing
            (OSPF Router ID, OspfRouter object) (key,value) pairs.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri_indices = {"name": parent_vrf.name, "version": cls.version}
        uri = cls.collection_uri.format(**uri_indices)
        try:
            response = session.request("GET", uri)
        except Exception as exc:
            raise ResponseError("GET", exc) from exc
        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)
        data = json.loads(response.text)
        ospf_dict = {}
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            # Create a OspfRouter object and adds it to parent Vrf object list
            instance_tag, ospf = cls.from_uri(session, parent_vrf, uri)
            # Load all OSPF Router data from within the Switch
            ospf_dict[instance_tag] = ospf

        return ospf_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create update an existing OSPF Router.
            Checks whether the VRF exists in the switch. Calls self.update() if
            OSPF Router is being updated. Calls self.create() if a new OSPF
            Router is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self._parent_vrf.materialized:
            self._parent_vrf.apply()
        if self.materialized:
            return self.update()
        return self.create()

    def __get_passive_interfaces_to_correct_form(self):
        """
        Auxiliary method to set passive interfaces to correct form for
            requests.
        """
        formatted_interfaces = {}
        if self.passive_interfaces is not None:
            for iface in self.passive_interfaces:
                if isinstance(iface, str):
                    iface = self.session.api.get_module(
                        self.session, "Interface", iface
                    )
                iface.get()
                if not iface.materialized:
                    raise VerificationError(
                        "Interface {0}".format(iface.name),
                        "Object inside passive_interfaces not materialized",
                    )
                formatted_iface = iface.get_info_format()
                formatted_interfaces.update(formatted_iface)
        return formatted_interfaces

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing OSPF Router.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        ospf_router_data = utils.get_attrs(self, self.config_attrs)

        formatted_ifaces = self.__get_passive_interfaces_to_correct_form()
        if formatted_ifaces != {}:
            ospf_router_data["passive_interfaces"] = formatted_ifaces

        self.__modified = self._put_data(ospf_router_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new  OSPF Router. Only returns if an
            exception is not raised.
        :return modified: True if entry was created.
        """
        ospf_router_data = utils.get_attrs(self, self.config_attrs)
        ospf_router_data["instance_tag"] = self.__instance_tag

        formatted_ifaces = self.__get_passive_interfaces_to_correct_form()
        if formatted_ifaces != {}:
            ospf_router_data["passive_interfaces"] = formatted_ifaces

        self.__modified = self._post_data(ospf_router_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete  OSPF Router table entry.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        self._parent_vrf.remove_ospf_router(self)
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vrf, response_data):
        """
        Create a OspfRouter object given a response_data related to the
            OspfRouter object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: Vrf object where OspfRouter object is stored.
        :param response_data: The response must be a dictionary of the form:
            {id: URL}, with the URL being of the form:
            "/rest/v10.04/system/vrfs/<name>/ospf_routers/<id>"
        :return: OspfRouter object.
        """
        ospf_arr = session.api.get_keys(response_data, cls.resource_uri_name)
        instance_tag = ospf_arr[0]
        return cls(session, instance_tag, parent_vrf)

    @classmethod
    def from_uri(cls, session, parent_vrf, uri):
        """
        Create a OspfRouter object given a URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return: tuple(OSPF Instance Tag: int, OSPF Router: OspfRouter).
        """
        # Obtain ID from URI like:
        # system/vrfs/{name}/ospf{version}_routers/{instance_tag}
        instance_tag = uri.split("/")[-1]

        # Create OspfRouter object
        ospf_obj = cls(session, instance_tag, parent_vrf)

        return instance_tag, ospf_obj

    def __str__(self):
        return "{0} with instance_tag {1}".format(
            type(self).__name__, self.instance_tag
        )

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific OSPF Router URI.
        return: Object's URI.
        """
        # PyaoscxModule's methods use self.path to store the URI
        return self.path

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def update_ospf_areas(self, new_area):
        """
        Update references to OSPF Areas. If an Area with the same area_id is
            found, update the reference to the new area, otherwise, add
            reference to the new area.
        :param new_area: Area to add reference to.
        """
        for area in self.areas:
            if area.area_id == new_area.area_id:
                # Make list element point to current object
                # See utils.list_attributes.ListDescriptor
                area = new_area
                return
        self.areas.append(new_area)

    def remove_ospf_area(self, area):
        """
        Update references to OSPF Areas. If an Area with the same area_id is
            found, delete the reference to it.
        :param new_area: Area to add reference to.
        """
        for area_ in self.areas:
            if area.area_id == area_.area_id:
                self.areas.remove(area_)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.pyaoscx_module import PyaoscxModule


class OspfArea(PyaoscxModule):
    """
    Provide configuration management for OSPF Area instance on AOS-CX devices.
    """

    collection_uri = (
        "system/vrfs/{name}/ospf{version}_routers/{instance_tag}/areas"
    )
    object_uri = collection_uri + "/{area_id}"

    indices = ["area_id"]
    resource_uri_name = "areas"

    # Use to manage references
    ospf_interfaces = ListDescriptor("ospf_interfaces")

    def __init__(self, session, area_id, parent_ospf_router, **kwargs):
        self.session = session
        # Assign ID
        self.__area_id = area_id
        # Assign parent OSPF Router
        self.__parent_ospf_router = parent_ospf_router
        # List used to determine attributes related to the OPSF configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Use to manage OSPF Interfaces
        self.ospf_interfaces = []
        # Attribute used to know if object was changed recently
        self.__modified = False
        self.base_uri = self.__parent_ospf_router.path + "/areas"
        self.path = "{0}/{1}".format(self.base_uri, self.__area_id)
        self.__parent_ospf_router.update_ospf_areas(self)

    @property
    def area_id(self):
        return self.__area_id

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a OSPF Area and fill the object
            with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is no exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)
        # Delete unwanted data
        if "ospf_interfaces" in data:
            data.pop("ospf_interfaces")
        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)
        # Determines if the OSPF Area is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", self.indices)
        # Set original attributes
        if "area_id" in data:
            del data["area_id"]
        self._original_attributes = data
        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        # Clean areas
        if self.ospf_interfaces == []:
            # Set Areas if any
            # Adds OSPF Interface to parent OSPF Area already
            from pyaoscx.ospf_interface import OspfInterface

            OspfInterface.get_all(self.session, self)

        return True

    @classmethod
    def get_all(cls, session, parent_ospf_router):
        """
        Perform a GET call to retrieve all system OSPF Area inside a OPSF
            Router, and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_ospf_router: parent OPSF Router object where OPSF Area
            is stored.
        :return: Dictionary containing OSPF Area IDs as keys and a OSPF Area
            objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri = "{0}/{1}/areas".format(
            parent_ospf_router.base_uri, parent_ospf_router.instance_tag
        )
        try:
            response = session.request("GET", uri)
        except Exception as exc:
            raise ResponseError("GET", exc) from exc
        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)
        data = json.loads(response.text)
        ospf_area_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            # Create an OspfArea object
            area_id, ospf_area = OspfArea.from_uri(
                session, parent_ospf_router, uri
            )
            # Load all OSPF Router data from within the Switch
            ospf_area_dict[area_id] = ospf_area
        return ospf_area_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Ospf Area.
            Checks whether the OSPF Area exists in the switch. Calls
            self.update() if OSPF Area is being updated. Calls self.create() if
            a new OSPF Area is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_ospf_router.materialized:
            self.__parent_ospf_router.apply()
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing OSPF Area.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # IMPORTANT: OSPF Area's ipsec_ah, ipsec_esp, and other_config MUST be
        # configured together, so if any of them needs to be updated, existing
        # values MUST be sent for all other attributes that have not changed
        # NOTE: 'other_config' is mistakenly not getting added to
        # self.config_attrs, so this is fixed here
        if "other_config" not in self.config_attrs:
            self.config_attrs.append("other_config")
        put_data = utils.get_attrs(self, self.config_attrs)
        self.__modified = self._put_data(put_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new OSPF Area. Only returns if an
            exception is not raised.
        :return modified: Boolean, True if entry was created.
        """
        post_data = utils.get_attrs(self, self.config_attrs)
        post_data["area_id"] = self.area_id

        self.__modified = self._post_data(post_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete OSPF Area table entry.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        self.__parent_ospf_router.remove_ospf_area(self)
        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_ospf_router, response_data):
        """
        Create an OspfArea object given a response_data related to the ospf
            router ID object
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_ospf_router: parent OspfRouter object where OspfArea
            object is stored.
        :param response_data: The response must be a dictionary of the form:
            { id: URL }, with the URL being of the form:
            "/rest/v10.04/system/vrfs/<name>/ospf_routers/<id>/areas/<id>"
        :return: OspfArea object.
        """
        ospf_area_arr = session.api.get_keys(
            response_data, cls.resource_uri_name
        )
        ospf_area_id = ospf_area_arr[0]
        return cls(session, ospf_area_id, parent_ospf_router)

    @classmethod
    def from_uri(cls, session, parent_ospf_router, uri):
        """
        Create an OspfArea object given a URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return index, ospf_area_obj: tuple containing both the OspfArea object
            and the OSPF Area's ID.
        """
        # Obtain ID from URI
        # system/vrfs/name/ospf_routers/1/areas/1.1.1.1
        # or
        # system/vrfs/name/ospfv3_routers/1/areas/1.1.1.1
        uri_parts = uri.split("/")
        area_id = uri_parts[-1]

        # Create OspfArea object
        ospf_area_obj = cls(
            session,
            area_id,
            parent_ospf_router,
        )

        return area_id, ospf_area_obj

    def __str__(self):
        return "{0} with area_id {1}".format(type(self).__name__, self.area_id)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific OSPF Area URI.
        return: Object's URI.
        """
        return self.path

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def update_ospf_interfaces(self, new_interface):
        """
        Update references to OSPF Interfaces. If an Interface with the same
            interface_name is found, update the reference to the new interface,
            otherwise, add reference to the new interface.
        :param new_interface: Interface to add reference to.
        """
        for interface in self.ospf_interfaces:
            if interface.interface_name == new_interface.interface_name:
                # Make list element point to current object
                # See utils.list_attributes.ListDescriptor
                interface = new_interface
                return
        self.ospf_interfaces.append(new_interface)

    def remove_ospf_interface(self, interface):
        """
        Update references to OSPF Interfaces. If an Interface with the same
            interface_name is found, delete the reference to it.
        :param interface: Interface to add reference to.
        """
        for i in self.ospf_interfaces:
            if i.interface_name == interface.interface_name:
                self.ospf_interfaces.remove(i)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.aggregate_address import AggregateAddress
from pyaoscx.bgp_neighbor import BgpNeighbor

from pyaoscx.pyaoscx_module import PyaoscxModule


class BgpRouter(PyaoscxModule):
    """
    Provide configuration management for BGP on AOS-CX devices.
    """

    indices = ["asn"]
    resource_uri_name = "bgp_routers"

    # Use to manage BGP Neighbors
    bgp_neighbors = ListDescriptor("bgp_neighbors")
    aggregate_addresses = ListDescriptor("aggregate_addresses")

    def __init__(self, session, asn: int, parent_vrf, uri=None, **kwargs):

        self.session = session
        # Assign id
        self.asn = asn
        # Assign parent Vrf object
        self.__set_vrf(parent_vrf)
        self._uri = uri
        # List used to determine attributes related to the BGP configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Use to manage BGP Neighbors
        self.bgp_neighbors = []
        # Use to manage Aggregate Addresses
        self.aggregate_addresses = []
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_vrf(self, parent_vrf):
        """
        Set parent Vrf object as an attribute for the BGP class.
        :param parent_vrf: a Vrf object.
        """
        # Set parent Vrf object
        self.__parent_vrf = parent_vrf

        # Set URI
        self.base_uri = "{0}/{1}/bgp_routers".format(
            self.__parent_vrf.base_uri,
            self.__parent_vrf.name,
        )

        # Verify BGP Router doesn't exist already inside VRF
        for bgp_router in self.__parent_vrf.bgp_routers:
            if bgp_router.asn == self.asn:
                # Make list element point to current object
                bgp_router = self
            else:
                # Add self to bgp_routers list in parent_vrf
                self.__parent_vrf.bgp_routers.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a BGP Router table entry and
            fill the object with the incoming attributes
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.asn)
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)
        # Delete unwanted data
        if "bgp_neighbors" in data:
            data.pop("bgp_neighbors")
        if "aggregate_addresses" in data:
            data.pop("aggregate_addresses")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the BGP Router is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self,
                data,
                "config_attrs",
                ["asn", "bgp_neighbors", "aggregate_addresses"],
            )

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "asn" in self.__original_attributes:
            self.__original_attributes.pop("asn")
        # Remove ID
        if "bgp_neighbors" in self.__original_attributes:
            self.__original_attributes.pop("bgp_neighbors")
        # Remove ID
        if "aggregate_addresses" in self.__original_attributes:
            self.__original_attributes.pop("aggregate_addresses")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        # Get BGP Neighbors
        if self.bgp_neighbors == []:
            # Set BGP Neighbor if any
            # Adds BGP Neighbor to parent BGP Router already
            BgpNeighbor.get_all(self.session, self)

        if self.aggregate_addresses == []:
            # Set Aggregate address if any
            # Adds Aggregate Addresses to parent BGP Router already
            AggregateAddress.get_all(self.session, self)
        return True

    @classmethod
    def get_all(cls, session, parent_vrf):
        """
        Perform a GET call to retrieve all system BGP inside a VRF, and create
            a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where VRF is stored.
        :return: Dictionary containing BGP Router IDs as keys and a BGP Router
            objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/bgp_routers".format(
            parent_vrf.base_uri, parent_vrf.name
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        bgp_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a BgpRouter object and adds it to parent Vrf object list
            asn, bgp = BgpRouter.from_uri(session, parent_vrf, uri)
            # Load all BGP Router data from within the Switch
            bgp.get()
            bgp_dict[asn] = bgp

        return bgp_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing BGP Router.
            Checks whether the BGP Router exists in the switch. Calls
            self.update() if BGP Router is being updated. Calls self.create()
            if a new BGP Router is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_vrf.materialized:
            self.__parent_vrf.apply()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing BGP Router.
        :return modified: True if Object was modified and a PUT request
            was made.
        """
        # Variable returned
        modified = False

        bgp_router_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(self.base_uri, self.asn)

        # Compare dictionaries
        if bgp_router_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(bgp_router_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = bgp_router_data

            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new BGP Router table entry. Only
            returns if no exception is raised.
        :return modified: Boolean, True if entry was created.
        """
        bgp_data = utils.get_attrs(self, self.config_attrs)
        bgp_data["asn"] = self.asn

        post_data = json.dumps(bgp_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        # Object was created, thus modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete BGP Router.
        """
        uri = "{0}/{1}".format(self.base_uri, self.asn)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from VRF
        for bgp_router in self.__parent_vrf.bgp_routers:
            if bgp_router.asn == self.asn:
                self.__parent_vrf.bgp_routers.remove(bgp_router)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vrf, response_data):
        """
        Create a BgpRouter object given a response_data related to it.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where VRF is stored.
        :param response_data: The response must be a dictionary of the form:
            {
                <asn>: "/rest/v10.04/system/vrfs/<name>/bgp_routers/<asn>"
            }
        :return: BgpRouter object.
        """
        bgp_arr = session.api.get_keys(
            response_data, BgpRouter.resource_uri_name
        )
        asn = bgp_arr[0]
        return BgpRouter(session, asn, parent_vrf)

    @classmethod
    def from_uri(cls, session, parent_vrf, uri):
        """
        Create a BgpRouter object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: parent Vrf object where BGP Router is stored.
        :param uri: a String with a URI.
        :return index, bgp_obj: tuple containing both the BgpRouter object
            and the BGP's asn.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)bgp_routers/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create BGP object
        bgp_obj = BgpRouter(session, index, parent_vrf, uri=uri)

        return index, bgp_obj

    def __str__(self):
        return "BGP Router ID {0}".format(self.asn)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific BGP Router URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.asn,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def create_bgp_neighbors(
        self,
        group_ip,
        family_type="l2vpn_evpn",
        reflector=False,
        send_community=False,
        local_interface="",
    ):
        """
        Perform a POST call to create BGP Neighbors to the associated current
            BGP Router. With l2vpn_evpn being True, this will also apply EVPN
            settings to the BGP neighbor configurations.
        :param group_ip: IPv4 address or name of group of the neighbors that
            functions as the BGP Router link. Example IPv4:
            10.10.12.11/255.255.255.255
        :param family_type: Alphanumeric to specify what type of neighbor
            settings to configure. The options are 'l2vpn-evpn',
            'ipv4-unicast', or 'ipv6-unicast'. When setting to l2vpn-evpn,
            the neighbor configurations also will add route-reflector-client
            and send-community settings. Defaults to "l2vpn_evpn".
        :param reflector: Boolean value to determine whether this neighbor has
            route reflector enabled. Defaults to False.
        :param send_community: Boolean value to determine whether this neighbor
            has send-community enabled. Defaults to False.
        :param local_interface: Optional alphanumeric to specify which
            interface the neighbor will apply to. Defaults to the empty string.
        :return bgp_neighbor_obj: BgpRouter object.
        """
        if not self.materialized:
            raise VerificationError(
                "VRF {0}".format(self.name), "Object not materialized"
            )

        if local_interface != "":
            if isinstance(local_interface, str):
                local_interface = self.session.api.get_module(
                    self.session, "Interface", local_interface
                )

        # Set values needed
        activate = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
            "l2vpn-evpn": False,
        }

        next_hop_unchanged = {"l2vpn-evpn": False}

        route_reflector_client = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
            "l2vpn-evpn": False,
        }

        send_community_data = {
            "ipv4-unicast": "none",
            "ipv6-unicast": "none",
            "l2vpn-evpn": "none",
        }

        activate[family_type] = True

        if send_community:
            send_community_data[family_type] = "both"

        if reflector:
            route_reflector_client[family_type] = reflector

        bgp_neighbor_obj = self.session.api.get_module(
            self.session,
            "BgpNeighbor",
            group_ip,
            parent_bgp_router=self,
            is_peer_group=False,
            remote_as=self.asn,
            shutdown=False,
            local_interface=local_interface,
            activate=activate,
            next_hop_unchanged=next_hop_unchanged,
            route_reflector_client=route_reflector_client,
            send_community=send_community_data,
        )

        # Try to obtain data; if not, create
        try:
            bgp_neighbor_obj.get()
        except GenericOperationError:
            # Create object inside switch
            bgp_neighbor_obj.apply()

        return bgp_neighbor_obj



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.ospf_router import OspfRouter


class Ospfv3Router(OspfRouter):
    """
    Provide configuration management for OSPFv3 Routers on AOS-CX devices.
    """

    version = "v3"

    resource_uri_name = "ospfv3_routers"



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.pyaoscx_error import PyaoscxError


class ResponseError(PyaoscxError):
    """
    Exception class for a PYAOSCX Response Error.
    """

    base_msg = "RESPONSE ERROR"

    def __init__(self, *args):
        self.message = None
        self.response = None
        if args:
            self.response = args[0]
            if len(args) > 1:
                self.message = ", ".join(str(a) for a in args[1:])
        else:
            self.message = None

    def __str__(self):
        msg_parts = [self.base_msg]
        if self.message:
            msg_parts.append(str(self.message))
        if self.response:
            msg_parts.append("Response")
            msg_parts.append(str(self.response))
        msg = ": ".join(msg_parts)
        return repr(msg)



from pyaoscx.session import Session
from pyaoscx.pyaoscx_factory import PyaoscxFactory
from pyaoscx.vlan import Vlan
from pyaoscx.interface import Interface

# There are two approaches to workflows, both using the session.
version = "10.04"
switch_ip = "172.25.0.2"
s = Session(switch_ip, version)
s.open("admin", "admin")

# Try block is used so that session closes even on error.
try:
    # APPROACH 1: OPEN GRANULATED APPROACH
    # VLAN
    # Create Vlan object -- Not yet materialized
    vlan100 = Vlan(s, 100, name="VLAN 100", voice=True)

    # Since object is not materialized, performs
    # a POST request -- This method internally
    # makes a GET request right after the POST
    # Obtaining all attributes VLAN related
    vlan100.apply()

    # Now let's create another object, that we know already exists
    # inside of the Switch
    vlan1 = Vlan(s, 1)
    # Perform a GET request to obtain all data and materialize object
    vlan1.get()
    # Now, we are able to modify the objects internal attributes
    vlan1.voice = True
    # Apply changes
    changed = vlan1.apply()
    # If changed is True, a PUT request was done and object was modified

    vlan100.description = "New description, changed via pyaoscx SDK"
    vlan100.apply()
    # Now vlan100 contains the description attribute
    print("VLAN 100 description {0}".format(vlan100.description))

    # ===========================================================
    # ===========================================================
    # ===========================================================

    # APPROACH 2: IMPERATIVE FACTORY APPROACH
    # VLAN
    # Create Factory object, passing the Session Object
    factory = PyaoscxFactory(s)

    # Create Vlan object
    # If vlan is non-existent, Factory instantly creates it
    # inside the switch device
    vlan200 = factory.vlan(200, "NAME200")

    # Now the granulated approach could still be used.
    # Or an imperative method too

    # ===========================================================
    # ===========================================================
    # ===========================================================

    # More complex example using the OPEN GRANULATED APPROACH
    # Create an Interface object

    lag = Interface(s, "lag1")
    lag.apply()

    # Create a Vlan object

    vlan_1 = Vlan(s, 1)
    # In this case, now that the VLAN exists within the Switch,
    # a GET request is called to obtain the VLAN's information.
    # The information is then added to the object as attributes.
    vlan_1.get()
    # Interfaces/Ports added to LAG
    port_1_1_11 = Interface(s, "1/1/11")
    port_1_1_11.get()
    # Make changes to configure LAG as L2
    lag.admin = "down"
    lag.routing = False
    lag.vlan_trunks = [vlan_1]
    lag.lacp = "passive"
    lag.other_config["mclag_enabled"] = False
    lag.other_config["lacp-fallback"] = False
    lag.vlan_mode = "native-untagged"
    lag.vlan_tag = vlan_1
    # Add port as LAG member
    lag.interfaces.append(port_1_1_11)

    # Apply changes
    lag.apply()

    # ===========================================================
    # ===========================================================
    # ===========================================================

    # Same complex example using the IMPERATIVE FACTORY APPROACH
    # PLUS USING IMPERATIVE METHODS

    # Create the Interface object
    lag2 = factory.interface("lag2")
    modified = lag2.configure_l2(
        description="Created using imperative method",
        admin="up",
        vlan_mode="native-untagged",
        vlan_tag=1,
        trunk_allowed_all=True,
        native_vlan_tag=True,
    )
    # If modified is True, a PUT request was done and object was modified

except Exception as error:
    print("Ran into exception: {0}. Closing session.".format(error))

finally:
    # At the end, the session MUST be closed
    s.close()



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.pyaoscx_error import PyaoscxError


class UnsupportedCapabilityError(PyaoscxError):
    """
    Exception class for an PYAOSCX Unsupported Capability Error.
    """

    base_msg = "UNSUPPORTED CAPABILITY"



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from urllib.parse import unquote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.ipv6 import Ipv6
from pyaoscx.pyaoscx_module import PyaoscxModule
from pyaoscx.vlan import Vlan
from pyaoscx.vrf import Vrf

from pyaoscx.interface import Interface as AbstractInterface


class Interface(AbstractInterface):
    """
    Provide configuration management for Interface and Ports for REST API
        Version 1. Uses methods inside AbstractInterface and any ones different
        are overridden by this class.
    """

    base_uri = "system/ports"
    base_uri_ports = "system/ports"
    base_uri_interface = "system/interfaces"

    indices = ["name"]

    ip6_addresses = ListDescriptor("ip6_addresses")

    def __init__(self, session, name, uri=None, **kwargs):
        self.session = session
        self._uri = None

        # List used to determine attributes related to the port configuration
        self.config_attrs = []
        # List used to determine attributes related to the interface
        # configuration
        self.config_attrs_int = []
        self.materialized = False

        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes_int = {}
        self.__original_attributes_port = {}

        # Set name, percents name and determine if Interface is a LAG
        self.__set_name(name)

        # List of previous interfaces before update
        # used to verify if a interface is deleted from lag
        self.__prev_interfaces = []

        # Use to manage IPv6 addresses
        self.ip6_addresses = []

        # Type required for configuration
        self.type = None
        # Set type
        self.__set_type()

        # Check if data should be added to object
        if self.__is_special_type:
            utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Port table entry, a Interface
            table entry and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.  The options are 'configuration', 'status', or 'statistics.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri_ports = "{0}/{1}".format(
            Interface.base_uri_ports, self.percents_name
        )

        # Bring Ports information
        try:
            response_ports = self.session.request(
                "GET", uri_ports, params=payload
            )

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response_ports, "GET"):
            raise GenericOperationError(
                response_ports.text, response_ports.status_code
            )

        data_port = json.loads(response_ports.text)
        # Adding ACL attributes to data_port
        # to then be added as attributes to the object
        acl_names = [
            "aclv6_in_cfg",
            "aclv4_in_cfg",
            "aclmac_in_cfg",
            "aclv4_out_cfg",
        ]

        for acl_attr in acl_names:
            data_port[acl_attr] = None

        # Delete ip6 addresses from incoming data
        data_port.pop("ip6_addresses")
        # Add Port dictionary as attributes for the object
        utils.create_attrs(self, data_port)

        # Determines if the module is configurable
        if selector in self.session.api.configurable_selectors:
            # Get list of keys and create a list with the given keys
            utils.set_config_attrs(
                self,
                data_port,
                "config_attrs",
                ["name", "origin", "other_config", "ip6_addresses"],
            )
        # Set original attributes
        self.__original_attributes_port = data_port
        # Delete unwanted attributes
        if "name" in self.__original_attributes_port:
            self.__original_attributes_port.pop("name")
        if "ip6_addresses" in self.__original_attributes_port:
            self.__original_attributes_port.pop("ip6_addresses")
        if "origin" in self.__original_attributes_port:
            self.__original_attributes_port.pop("origin")
        if "other_config" in self.__original_attributes_port:
            self.__original_attributes_port.pop("other_config")

        # Check if port is a LAG
        # If not, makes get request to system/interfaces
        if self.type != "lag":
            uri_interfaces = "{0}/{1}".format(
                Interface.base_uri_interface, self.percents_name
            )
            # Bring Interface information
            try:
                response_ints = self.session.request(
                    "GET", uri_interfaces, params=payload
                )

            except Exception as e:
                raise ResponseError("GET", e)

            if not utils._response_ok(response_ints, "GET"):
                raise GenericOperationError(
                    response_ints.text, response_ints.status_code
                )

            data_int = json.loads(response_ints.text)
            # Add Interface dictionary as attributes for the object
            utils.create_attrs(self, data_int)

            # Determines if the module is configurable
            if selector in self.session.api.configurable_selectors:
                # Get list of keys and create a list with the given keys
                utils.set_config_attrs(
                    self, data_int, "config_attrs_int", ["name", "origin"]
                )
            # Set original attributes
            self.__original_attributes_int = data_int
            # Delete unwanted attributes
            if "name" in self.__original_attributes_int:
                self.__original_attributes_int.pop("name")
            if "origin" in self.__original_attributes_int:
                self.__original_attributes_int.pop("origin")

        # Set a list of interfaces as an attribute
        if hasattr(self, "interfaces") and self.interfaces is not None:
            interfaces_list = []
            # Get all URI elements in the form of a list
            uri_list = self.session.api.get_uri_from_data(self.interfaces)
            for uri in uri_list:
                # Create an Interface object
                name, interface = Interface.from_uri(self.session, uri)

                # Check for circular reference
                # No need to get() if it's circular; it is already
                # materialized. Just set flag
                if name == self.name:
                    interface.materialized = True
                else:
                    # Materialize interface
                    interface.get()

                # Add interface to list
                interfaces_list.append(interface)

            # Set list as Interfaces
            self.interfaces = interfaces_list
            # Set list of interfaces
            self.__prev_interfaces = list(self.interfaces)

        # Set VRF
        if hasattr(self, "vrf") and self.vrf is not None:
            # Set keepalive VRF as a Vrf object
            vrf_obj = Vrf.from_response(self.session, self.vrf)
            self.vrf = vrf_obj
            # Materialized VRF
            self.vrf.get()

        # Set VLAN
        if hasattr(self, "vlan_tag") and self.vlan_tag is not None:
            # Set VLAN as a Vlan Object
            vlan_obj = Vlan.from_response(self.session, self.vlan_tag)
            self.vlan_tag = vlan_obj
            # Materialized VLAN
            self.vlan_tag.get()

        # vlan_trunks
        # Set a list of vlans as an attribute
        if hasattr(self, "vlan_trunks") and self.vlan_trunks is not None:
            vlan_trunks = []
            # Get all URI elements in the form of a list
            uri_list = self.session.api.get_uri_from_data(self.vlan_trunks)

            for uri in uri_list:
                # Create a Vlan object
                vlan_id, vlan = Vlan.from_uri(self.session, uri)
                # Materialize VLAN
                vlan.get()
                # Add VLAN to dictionary
                vlan_trunks.append(vlan)
            # Set list as VLANs
            self.vlan_trunks = vlan_trunks

        # Set all ACLs
        from pyaoscx.acl import ACL

        if hasattr(self, "aclmac_in_cfg") and self.aclmac_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclmac_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclmac_in_cfg = acl

        if hasattr(self, "aclmac_out_cfg") and self.aclmac_out_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclmac_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclmac_out_cfg = acl

        if hasattr(self, "aclv4_in_cfg") and self.aclv4_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_in_cfg = acl

        if hasattr(self, "aclv4_out_cfg") and self.aclv4_out_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_out_cfg = acl

        if hasattr(self, "aclv4_routed_in_cfg") and self.aclv4_routed_in_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_routed_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_routed_in_cfg = acl

        if hasattr(self, "aclv4_routed_out_cfg") and self.aclv4_routed_out_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_routed_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_routed_out_cfg = acl

        if hasattr(self, "aclv6_in_cfg") and self.aclv6_in_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_in_cfg = acl

        if hasattr(self, "aclv6_out_cfg") and self.aclv6_out_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_out_cfg = acl

        if hasattr(self, "aclv6_routed_in_cfg") and self.aclv6_routed_in_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_routed_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_routed_in_cfg = acl

        if hasattr(self, "aclv6_routed_out_cfg") and self.aclv6_routed_out_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_routed_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_routed_out_cfg = acl

        # Set IPv6 addresses if any
        if self.ip6_addresses == []:
            # Set IPv6 addresses if any
            # Loads IPv6 already into the Interface
            Ipv6.get_all(self.session, self)

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system Ports and return a list of
            them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing ports IDs as keys and a port objects
            as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", Interface.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        interfaces_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a Interface object
            name, interface = Interface.from_uri(session, uri)

            interfaces_dict[name] = interface

        return interfaces_dict

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create an Interface object given a interface URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return name, Interface: tuple containing both the interface's name
            and an Interface object.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)/(?P<index>.+)")
        name_percents = index_pattern.match(uri).group("index")
        name = unquote_plus(name_percents)
        # Create Interface object
        interface_obj = Interface(session, name, uri=uri)

        return name, interface_obj

    @classmethod
    def get_facts(cls, session):
        """
        Perform a GET call to retrieve all Interfaces and their respective
            data.
        :param cls: Class reference.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return facts: Dictionary containing Interface IDs as keys and
            Interface objects as values.
        """
        logging.info("Retrieving the switch interfaces facts")

        depth = session.api.default_facts_depth

        payload = {"depth": depth}

        # Get Ports information
        try:
            response_ports = session.request(
                "GET", Interface.base_uri_ports, params=payload
            )

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response_ports, "GET"):
            raise GenericOperationError(
                response_ports.text, response_ports.status_code
            )

        ports_data = json.loads(response_ports.text)

        # Build interface URI
        uri_interface = "{0}?depth={1}".format(
            Interface.base_uri_interface, depth
        )

        # Get Interface information
        try:
            response_interface = session.request(
                "GET", uri_interface, params=payload
            )

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response_interface, "GET"):
            raise GenericOperationError(
                response_interface.text, response_interface.status_code
            )

        # Load response text into json format
        interfaces_data = json.loads(response_interface.text)
        facts_dict = {}

        # Merge Ports and Interfaces by key name
        for port in ports_data:
            if "name" in port:
                facts_dict[port["name"]] = port
        for interface in interfaces_data:
            if "name" in interface:
                if interface["name"] in facts_dict:
                    facts_dict[interface["name"]].update(interface)
                else:
                    facts_dict[interface["name"]] = interface

        return facts_dict

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create an Interface object given a response_data related to the
            Interface object.
        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param response_data: The response must be a dictionary of the form:
            {id: URL}, with the URL being of the form:
            "/rest/v1/system/interfaces/1"
            or
            "/rest/v1/system/ports/1"
        :return: Interface object.
        """
        try:
            # Try using interfaces
            interfaces_id_arr = session.api.get_keys(
                response_data, "interfaces"
            )
        except AttributeError:
            # If AttributeError for Nonetype, try with ports
            interfaces_id_arr = session.api.get_keys(response_data, "ports")
        interface_name = interfaces_id_arr[0]
        return session.api.get_module(session, "Interface", interface_name)

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete Interface.
        """
        if not self.__is_special_type:
            raise VerificationError("Interface", "Can't be deleted")

        # Delete Interface via a DELETE request to Ports Table
        uri = "{0}/{1}".format(Interface.base_uri_ports, self.name)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        # DELETE Interface via DELETE request to Interface Table
        # Check if port is a LAG
        # If not, DELETE request to Interface Table
        if self.type != "lag":
            try:
                response_ints = self.session.request(
                    "DELETE", Interface.base_uri_interface
                )

            except Exception as e:
                raise ResponseError("DELETE", e)

            if not utils._response_ok(response_ints, "DELETE"):
                raise GenericOperationError(
                    response_ints.text, response_ints.status_code
                )

        # Clean LAG from interfaces
        # Delete interface references
        for interface in self.__prev_interfaces:
            # If interface name is not the same as the current one
            if interface.name != self.name and self.type == "lag":
                interface.__delete_lag(self)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to update data for a Port and Interface table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Flag used to determine if Object was modified
        modified_port = True
        modified_int = True
        # Check if Object is a LAG
        if self.type != "lag":
            uri_interfaces = "{0}/{1}".format(
                Interface.base_uri_interface, self.percents_name
            )
            # get Interface data related to configuration
            int_data = utils.get_attrs(self, self.config_attrs_int)
            # Remove type
            if "type" in int_data:
                int_data.pop("type")
            if "type" in self.__original_attributes_int:
                self.__original_attributes_int.pop("type")
            # Set put_int_data
            put_int_data = json.dumps(int_data)
            # Compare dictionaries
            if int_data == self.__original_attributes_int:
                # Object was not modified
                modified_port = False
            else:
                # Bring Interface information
                try:
                    response_ints = self.session.request(
                        "PUT", uri_interfaces, data=put_int_data
                    )

                except Exception as e:
                    raise ResponseError("PUT", e)

                if not utils._response_ok(response_ints, "PUT"):
                    raise GenericOperationError(
                        response_ints.text, response_ints.status_code
                    )

                # Set new original attributes
                self.__original_attributes_int = int_data

        uri_ports = "{0}/{1}".format(
            Interface.base_uri_ports, self.percents_name
        )

        # get Port data related to configuration
        port_data = utils.get_attrs(self, self.config_attrs)

        # Check for Ipv4
        try:
            if self.ip4_address is not None:
                port_data["ip4_address"] = self.ip4_address
        except AttributeError:
            pass
        # Check if vrf is inside the data related to Port
        if "vrf" in port_data:
            # Set VRF in the correct format for PUT
            port_data["vrf"] = self.vrf.get_info_format()

        # Check if vlan_tag is inside the data related to Port
        if "vlan_tag" in port_data:
            # Set VLAN in the correct format for PUT
            port_data["vlan_tag"] = self.vlan_tag.get_info_format()

        # Set interfaces into correct form
        if "interfaces" in port_data:
            # Check for interfaces no longer in LAG
            if self.__is_special_type and self.type == "lag":
                for element in self.__prev_interfaces:
                    # If element was deleted from interfaces
                    if element not in self.interfaces:
                        # Delete element reference to current LAG
                        element.__delete_lag(self)

            # Set prev interfaces with current ones
            # Copies interfaces
            self.__prev_interfaces = list(self.interfaces)

            formated_interfaces = []
            # Set interfaces into correct form
            for element in self.interfaces:
                # If element is the same as current, ignore
                if element.name == self.name and self.type == "lag":
                    pass
                else:
                    # Verify object is materialized
                    if not element.materialized:
                        raise VerificationError(
                            "Interface {0}".format(element.name),
                            "Object inside interfaces not materialized",
                        )
                    # Only in V1 get_uri() is used,
                    # In any other version, element.get_info_format()
                    # is used
                    formated_element = element.get_uri(True)
                    formated_interfaces.append(formated_element)

                    if self.type == "lag":
                        # New element being added to LAG
                        element.__add_member_to_lag(self)

            # Set values in correct form
            port_data["interfaces"] = formated_interfaces

        # Set VLANs into correct form
        if "vlan_trunks" in port_data:
            formated_vlans = []
            # Set interfaces into correct form
            for element in self.vlan_trunks:
                # Verify object is materialized
                if not element.materialized:
                    raise VerificationError(
                        "Vlan {0}".format(element),
                        "Object inside vlan trunks not materialized",
                    )
                formated_element = element.get_info_format()
                formated_vlans.append(formated_element)

            # Set values in correct form
            port_data["vlan_trunks"] = formated_vlans

        # Set all ACLs
        if "aclmac_in_cfg" in port_data and self.aclmac_in_cfg is not None:
            # Set values in correct form
            port_data["aclmac_in_cfg"] = self.aclmac_in_cfg.get_info_format()

        if "aclmac_out_cfg" in port_data and self.aclmac_out_cfg is not None:
            # Set values in correct form
            port_data["aclmac_out_cfg"] = self.aclmac_out_cfg.get_info_format()

        if "aclv4_in_cfg" in port_data and self.aclv4_in_cfg is not None:
            # Set values in correct form
            port_data["aclv4_in_cfg"] = self.aclv4_in_cfg.get_info_format()

        if "aclv4_out_cfg" in port_data and self.aclv4_out_cfg is not None:
            # Set values in correct form
            port_data["aclv4_out_cfg"] = self.aclv4_out_cfg.get_info_format()

        if "aclv4_routed_in_cfg" in port_data and self.aclv4_routed_in_cfg:
            # Set values in correct form
            port_data[
                "aclv4_routed_in_cfg"
            ] = self.aclv4_routed_in_cfg.get_info_format()

        if "aclv4_routed_out_cfg" in port_data and self.aclv4_routed_out_cfg:
            # Set values in correct form
            port_data[
                "aclv4_routed_out_cfg"
            ] = self.aclv4_routed_out_cfg.get_info_format()

        if "aclv6_in_cfg" in port_data and self.aclv6_in_cfg is not None:
            # Set values in correct form
            port_data["aclv6_in_cfg"] = self.aclv6_in_cfg.get_info_format()

        if "aclv6_out_cfg" in port_data and self.aclv6_out_cfg is not None:
            # Set values in correct form
            port_data["aclv6_out_cfg"] = self.aclv6_out_cfg.get_info_format()

        if "aclv6_routed_in_cfg" in port_data and self.aclv6_routed_in_cfg:
            # Set values in correct form
            port_data[
                "aclv6_routed_in_cfg"
            ] = self.aclv6_routed_in_cfg.get_info_format()

        if "aclv6_routed_out_cfg" in port_data and self.aclv6_routed_out_cfg:
            # Set values in correct form
            port_data[
                "aclv6_routed_out_cfg"
            ] = self.aclv6_routed_out_cfg.get_info_format()

        # Set addresses the correct way
        if self.ip6_addresses is not None:
            ip6_addresses_dict = {}

            for ip in self.ip6_addresses:
                ip6_addresses_dict[ip.address] = ip.get_uri()

            # Set values in correct form
            port_data["ip6_addresses"] = ip6_addresses_dict

        # Delete type from Port data
        if "type" in port_data:
            port_data.pop("type")

        if "type" in self.__original_attributes_port:
            self.__original_attributes_port.pop("type")
        # Special case, if dictionary is empty
        if port_data["ip6_addresses"] == {}:
            self.__original_attributes_port["ip6_addresses"] = {}

        # Compare dictionaries
        if port_data == self.__original_attributes_port:
            # Object was not modified
            modified_int = False
        else:
            # Set put_port_data
            put_port_data = json.dumps(port_data)

            # Bring Port information
            try:
                response_ports = self.session.request(
                    "PUT", uri_ports, data=put_port_data
                )

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response_ports, "PUT"):
                raise GenericOperationError(
                    response_ports.text, response_ports.status_code
                )
            # Set new original attributes
            self.__original_attributes_port = port_data

        return modified_int or modified_port

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a Port table entry for Interface. Only
            returns if no exception is raised.
        :return True if entry was created inside Device.
        """
        port_data = utils.get_attrs(self, self.config_attrs)
        port_data["name"] = self.name

        post_data_ports = json.dumps(port_data)
        try:
            response = self.session.request(
                "POST", Interface.base_uri_ports, data=post_data_ports
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Check if port is a LAG
        # If not, POST Request to Interface Table
        if self.type != "lag":
            # Set data for Interface Table
            interface_data = utils.get_attrs(self, self.config_attrs_int)
            interface_data["name"] = self.name
            interface_data["type"] = self.type
            interface_data["referenced_by"] = self.get_uri()

            # Set post_int_data
            post_int_data = json.dumps(interface_data)

            # Bring Interface information
            try:
                response_ints = self.session.request(
                    "POST", Interface.base_uri_interface, data=post_int_data
                )

            except Exception as e:
                raise ResponseError("POST", e)

            if not utils._response_ok(response_ints, "POST"):
                raise GenericOperationError(
                    response_ints.text, response_ints.status_code
                )

        # Get all objects data
        self.get()
        # Object was created
        return True

    @PyaoscxModule.deprecated
    def get_uri(self, interface=False):
        """
        Method used to obtain the specific Interface URI.
        :param interface: Boolean, if true URI would contain
            a /interfaces/interface_name instead of
            /ports/interface_name.
        """
        uri = ""
        if not interface:
            uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                Interface.base_uri_ports,
                self.percents_name,
            )
        else:
            uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                Interface.base_uri_interface,
                self.percents_name,
            )

        return uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    def __str__(self):
        return "Port name: '{0}'".format(self.name)

    def __set_to_default(self):
        """
        Perform a PUT call to set Interface to default settings.
        :return: True if object was changed.
        """
        uri_ports = "{0}/{1}".format(
            Interface.base_uri_ports, self.percents_name
        )

        # get Port data related to configuration
        port_data = {}

        # Set interfaces into correct form
        if "interfaces" in port_data:
            if self.__is_special_type and self.name == "lag":
                self.interfaces = []
                for element in self.__prev_interfaces:
                    # If element was deleted from interfaces
                    if element not in self.interfaces:
                        # Delete element reference to current LAG
                        element.__delete_lag(self)
            else:
                self.interfaces = [self]

            # Set prev interfaces with current ones
            # Copies interfaces
            self.__prev_interfaces = list(self.interfaces)

        # Set put_port_data
        put_port_data = json.dumps(port_data)

        # Update Port information
        try:
            response_ports = self.session.request(
                "PUT", uri_ports, data=put_port_data
            )

        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response_ports, "PUT"):
            raise GenericOperationError(
                response_ports.text, response_ports.status_code
            )

        # Update values with new ones
        self.get()
        return True



# (C) Copyright 2019-2021 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import re

from datetime import date

from pyaoscx.api import API


class v1(API):
    """
    Represents a REST API Version 1. It keeps all the information needed for
        the version and methods related to it.
    """

    def __init__(self):
        self.release_date = date(2017, 1, 1)
        self.version = "1"
        self.default_selector = "configuration"
        self.default_depth = 0
        self.default_facts_depth = 1
        self.default_subsystem_facts_depth = 4
        self.valid_selectors = ["configuration", "status", "statistics"]
        self.configurable_selectors = ["configuration"]
        self.compound_index_separator = "/"
        self.valid_depths = [0, 1, 2, 3]

    def get_index(self, obj):
        """
        Method used to obtain the correct format of the objects information
            which depends on the Current API version.
        :param obj: PyaoscxModule object.
        :return: Resource URI.
        """
        # use object indices
        return obj.get_uri()

    def get_keys(self, response_data, module_name):
        """
        Given a string obtain the keys in it and return them.
        :param response_data: a string of the form:
            "/rest/v1/system/<module>/<key_1>/<key_2>".
        :return name_arr: List of keys.
        """
        # Create regex string
        regex_str = r"(.*)/" + re.escape(module_name) + r"/(?P<ids>.+)"
        # Pattern expected
        ids_pattern = re.compile(regex_str)
        # Match pattern
        indices = ids_pattern.match(response_data).group("ids")
        # Get all indices
        indices = indices.split("/")

        return indices

    def get_uri_from_data(self, data):
        """
        Given a response data, create a list of URI items. In this Version the
            data is a list, string or dict.
        :param data: String, List or Dictionary containing URI items.
        :return uri_list: Return the list of URIs.
        """
        if isinstance(data, list):
            return data
        elif isinstance(data, str):
            return [data]
        elif isinstance(data, dict):
            uri_list = []
            for k, v in data.items():
                item = self.get_uri_from_data(v)
                # if value is a list, then concatenate.
                if isinstance(item, list):
                    uri_list += item
                elif isinstance(item, str):
                    uri_list.append(item)

            return uri_list

    def _create_ospf_area(self, module_class, session, index_id, **kwargs):
        return module_class(session, index_id, **kwargs)



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from urllib.parse import quote_plus, unquote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class AggregateAddress(PyaoscxModule):
    """
    Provide configuration management for Aggregate Address on AOS-CX devices.
    """

    indices = ["address-family", "ip_prefix"]
    resource_uri_name = "aggregate_addresses"

    def __init__(
        self,
        session,
        address_family,
        ip_prefix,
        parent_bgp_router,
        uri=None,
        **kwargs
    ):
        self.session = session
        # Assign ID
        self.address_family = address_family
        self.__set_name(ip_prefix)
        # Assign parent BGP Router
        self.__set_parent_bgp_router(parent_bgp_router)
        self._uri = uri
        # List used to determine attributes related to the Aggregate Address
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_name(self, ip_prefix):
        """
        Set name attribute in the proper form for references.
        :param ip_prefix: Object's IP.
        """
        # Add attributes to class
        self._is_lag = False
        self.ip_prefix = None
        self.reference_ip_prefix = None
        if r"%2F" in ip_prefix:
            self.ip_prefix = unquote_plus(ip_prefix)
            self.reference_ip_prefix = ip_prefix
        else:
            self.ip_prefix = ip_prefix
            self.reference_ip_prefix = quote_plus(self.ip_prefix)

    def __set_parent_bgp_router(self, parent_bgp_router):
        """
        Set parent BgpRouter as an attribute for the Aggregate Address class.
        :param parent_bgp_router a BgpRouter object.
        """
        # Set parent BGP Router
        self.__parent_bgp_router = parent_bgp_router

        # Set URI
        self.base_uri = "{0}/{1}/aggregate_addresses".format(
            self.__parent_bgp_router.base_uri, self.__parent_bgp_router.asn
        )

        for address in self.__parent_bgp_router.aggregate_addresses:
            if (
                address.address_family == self.address_family
                and address.ip_prefix == self.ip_prefix
            ):
                # Make list element point to current object
                address = self
            else:
                # Add self to Aggregate Addresses list in parent BGP Router
                self.__parent_bgp_router.aggregate_addresses.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Aggregate Address table entry
            and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}{2}{3}".format(
            self.base_uri,
            self.address_family,
            self.session.api.compound_index_separator,
            self.reference_ip_prefix,
        )

        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the Aggregate Address is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["address-family"]
            )

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "address-family" in self.__original_attributes:
            self.__original_attributes.pop("address-family")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_bgp_router):
        """
        Perform a GET call to retrieve all system Aggregate Addresses inside a
            BGP Router, and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_bgp_router: parent bgp_router object where Aggregate
            Address is stored.
        :return: Dictionary containing Aggregate Addresses IDs as keys and a
            AggregateAddress objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/aggregate_addresses".format(
            parent_bgp_router.base_uri, parent_bgp_router.asn
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        agg_address_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a AggregateAddress object
            indices, aggregate_address = AggregateAddress.from_uri(
                session, parent_bgp_router, uri
            )
            agg_address_dict[indices] = aggregate_address

        return agg_address_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method to either create or update an existing Aggregate Address.
            Checks whether the Aggregate Addresses exists in the switch. Calls
            self.update() if Aggregate Address is being updated. Calls
            self.create() if a new Aggregate Address is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_bgp_router.materialized:
            self.__parent_bgp_router.apply()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Aggregate Address.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        agg_address_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}{2}{3}".format(
            self.base_uri,
            self.address_family,
            self.session.api.compound_index_separator,
            self.reference_ip_prefix,
        )
        # Compare dictionaries
        if agg_address_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(agg_address_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)
            # Set new original attributes
            self.__original_attributes = agg_address_data

            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Aggregate Address table entry. Only
            returns if no exception is raised.
        :return modified: True if entry was created.
        """
        ag_address_data = utils.get_attrs(self, self.config_attrs)
        ag_address_data["address-family"] = self.address_family
        ag_address_data["ip_prefix"] = self.ip_prefix

        post_data = json.dumps(ag_address_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        # Object was modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete Aggregate Address.
        """
        uri = "{0}/{1}{2}{3}".format(
            self.base_uri,
            self.address_family,
            self.session.api.compound_index_separator,
            self.reference_ip_prefix,
        )

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from BGP Router
        for address in self.__parent_bgp_router.aggregate_addresses:
            if (
                address.address_family == self.address_family
                and address.ip_prefix == self.ip_prefix
            ):
                self.__parent_bgp_router.aggregate_addresses.remove(address)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_bgp_router, response_data):
        """
        Create a AggregateAddress object given a response_data related to the
            Aggregate Address ID object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_bgp_router: parent BGP Router class where Aggregate
            Address is stored.
        :param response_data: The response must be a dictionary of the form:
            {id: URL}, with the URL being of the form:
            "bgp_routers/<asn>/aggregate_addresses/<id>"
            under the path:
            "/rest/v10.04/system/vrfs/<vrf_name>/"
        :return: AggregateAddress object.
        """
        aggr_address_arr = session.api.get_keys(
            response_data, cls.resource_uri_name
        )
        ip_prefix = aggr_address_arr[1]
        address_family = aggr_address_arr[0]

        return AggregateAddress(
            session, address_family, ip_prefix, parent_bgp_router
        )

    @classmethod
    def from_uri(cls, session, parent_bgp_router, uri):
        """
        Create a AggregateAddress object.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_bgp_router: parent BGP Router class where Aggregate
            Address is stored.
        :param uri: a String with a URI.
        :return indices, aggr_address: tuple containing both the Aggregate
            Address object and the Aggregate Address' ID.
        """
        # Obtain ID from URI
        index_pattern = re.compile(
            r"(.*)aggregate_addresses/(?P<index1>.+)[,./-](?P<index2>.+)"
        )
        index1 = index_pattern.match(uri).group("index1")
        index2 = index_pattern.match(uri).group("index2")

        # Create Create a AggregateAddress object
        aggr_address = AggregateAddress(
            session, index1, index2, parent_bgp_router
        )
        indices = "{0},{1}".format(index1, index2)

        return indices, aggr_address

    def __str__(self):
        return "Aggregate Address ID {0}".format(self.address_family)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific Aggregate Address URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}{3}{4}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.address_family,
                self.session.api.compound_index_separator,
                self.reference_ip_prefix,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.verification_error import VerificationError


class ParameterError(VerificationError):
    """
    Exception raised when wrong parameters are passed to functions.
    """

    base_msg = "PARAMETER ERROR"



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import random
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.mac import Mac
from pyaoscx.static_mac import StaticMac

from pyaoscx.pyaoscx_module import PyaoscxModule


class Vlan(PyaoscxModule):
    """
    Provide configuration management for VLANs on AOS-CX devices.
    """

    base_uri = "system/vlans"
    resource_uri_name = "vlans"
    indices = ["id"]

    macs = ListDescriptor("macs")
    static_macs = ListDescriptor("static_macs")

    def __init__(self, session, vlan_id, uri=None, **kwargs):

        self.session = session
        self._uri = uri
        self.id = vlan_id
        # List used to determine attributes related to the VLAN configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        utils.set_creation_attrs(self, **kwargs)
        # Use to manage MACs
        self.macs = []
        # Use to manage Static MACs
        self.static_macs = []
        # Attribute used to know if object was changed recently
        self.__modified = False

        # Build the path that identifies the current Vlan
        self.path = "{0}/{1}".format(Vlan.base_uri, self.id)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a VLAN table entry and fill the
            object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Delete unwanted data
        if "macs" in data:
            data.pop("macs")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the VLAN is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", ["id", "macs"])

        # Set original attributes
        self._original_attributes = data
        # Remove ID
        if "id" in self._original_attributes:
            self._original_attributes.pop("id")
        # Remove macs
        if "macs" in self._original_attributes:
            self._original_attributes.pop("macs")

        # Set all ACLs
        from pyaoscx.acl import ACL

        if hasattr(self, "aclmac_in_cfg") and self.aclmac_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclmac_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclmac_in_cfg = acl

        if hasattr(self, "aclv4_in_cfg") and self.aclv4_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_in_cfg = acl

        if hasattr(self, "aclv6_in_cfg") and self.aclv6_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_in_cfg = acl

        # Clean MACs
        if not self.macs:
            # Set MACs if any
            # Adds macs to parent Vlan object
            Mac.get_all(self.session, self)

        # Set Static MACs
        if self.static_macs == []:
            StaticMac.get_all(self.session, self)

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system VLAN and create a dictionary
            containing each respective VLAN.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing VLAN IDs as keys and a Vlan object as
            value.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", Vlan.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        vlans_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a Vlan object
            vlan_id, vlan = Vlan.from_uri(session, uri)

            vlans_dict[vlan_id] = vlan

        return vlans_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create a VLAN, or update an existing one.
            Checks whether the VLAN exists in the switch. Calls self.update()
            if VLAN is being updated. Calls self.create() if the VLAN doesn't
            exist in the switch.
        :return modified: Boolean, True if object was created or modified.
        """
        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing VLAN table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        vlan_data = utils.get_attrs(self, self.config_attrs)

        # Set all ACLs
        if "aclmac_in_cfg" in vlan_data and self.aclmac_in_cfg is not None:
            # Set values in correct form
            vlan_data["aclmac_in_cfg"] = self.aclmac_in_cfg.get_info_format()

        if "aclv4_in_cfg" in vlan_data and self.aclv4_in_cfg is not None:
            # Set values in correct form
            vlan_data["aclv4_in_cfg"] = self.aclv4_in_cfg.get_info_format()

        if "aclv6_in_cfg" in vlan_data and self.aclv6_in_cfg is not None:
            # Set values in correct form
            vlan_data["aclv6_in_cfg"] = self.aclv6_in_cfg.get_info_format()

        return self._put_data(vlan_data)

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new VLAN using the object's attributes
            as POST body. Exception is raised if object is unable to be
            created.
        :return modified: Boolean, True if VLAN was created.
        """
        # Get all VLAN data given by the user
        vlan_data = utils.get_attrs(self, self.config_attrs)
        if isinstance(self.id, str):
            self.id = int(self.id)
        vlan_data["id"] = self.id

        return self._post_data(vlan_data)

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete VLAN table entry.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a Vlan object given a response_data related to the Vlan object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                1: "/rest/v10.04/system/vlans/1"
            }
        :return: Vlan Object.
        """
        vlan_id_arr = session.api.get_keys(
            response_data, Vlan.resource_uri_name
        )
        vlan_id = vlan_id_arr[0]
        return Vlan(session, vlan_id)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Vlan object given a VLAN URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return vlan_id, vlan: tuple with the Vlan object its ID.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)vlans/(?P<index>.+)")
        index_str = index_pattern.match(uri).group("index")
        vlan_id = int(index_str)
        # Create Vlan object
        vlan_obj = Vlan(session, vlan_id, uri=uri)

        return vlan_id, vlan_obj

    @classmethod
    def get_facts(cls, session):
        """
        Modify this to Perform a GET call to retrieve all VLANs and their
            respective data.
        :param cls: Class reference.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return facts: Dictionary containing VLAN IDs as keys and Vlan objects
            as values.
        """
        logging.info("Retrieving switch VLANs facts")

        # Set VLAN facts depth
        vlan_depth = session.api.default_facts_depth

        # Build URI
        uri = "{0}?depth={1}".format(Vlan.base_uri, vlan_depth)

        try:
            # Try to get facts data via GET method
            response = session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)
        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Load response text into json format
        facts = json.loads(response.text)

        # Delete internal VLANs
        internal_vlan_list = []
        for vlan in facts.keys():
            if "type" in facts[vlan].keys():
                if facts[vlan]["type"] == "internal":
                    internal_vlan_list.append(vlan)

        for vlan in internal_vlan_list:
            facts.pop(vlan)

        return facts

    def __str__(self):
        try:
            return "Vlan, name: '{0}' ID: '{1}' and description: '{2}'".format(
                self.name, self.id, self.description
            )
        except Exception:
            return "Vlan, ID: '{0}'".format(self.id)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific VLAN URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix, Vlan.base_uri, self.id
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def modify(self, vlan_name=None, vlan_desc=None, admin_conf_state=None):
        """
        Perform a PUT calls to modify an existing VLAN.
        :param vlan_name: Optional Alphanumeric name of VLAN. Won't be modified
            if not specified.
        :param vlan_desc: Optional description to add to VLAN. Won't be
            modified if not specified.
        :param admin_conf_state: Optional administratively-configured state of
            VLAN. Won't be modified if not specified. Only configurable for
            static VLANs.
        :return: True if object was changed.
        """
        if vlan_name is not None:
            self.name = vlan_name

        if vlan_desc is not None:
            self.description = vlan_desc

        if self.type == "static" and admin_conf_state is not None:
            # admin-configured state can only be set on static VLANs
            self.admin = admin_conf_state

        # Apply changes inside switch
        return self.apply()

    def attach_acl_in(self, acl_name, list_type):
        """
        Update ACL IN values inside a Vlan object.
        :param acl_name: Alphanumeric String that is the name of the ACL.
        :param list_type: Alphanumeric String of ipv4, ipv6, or mac to specify
            the type of ACL.
        :return: True if object was changed.
        """
        # Create Acl object
        acl_obj = self.session.api.get_module(
            self.session, "ACL", index_id=acl_name, list_type=list_type
        )

        if list_type == "ipv6":
            self.aclv6_in_cfg = acl_obj
            if self.aclv6_in_cfg_version is None:
                self.aclv6_in_cfg_version = random.randint(
                    -9007199254740991, 9007199254740991
                )
        if list_type == "ipv4":
            self.aclv4_in_cfg = acl_obj
            if self.aclv4_in_cfg_version is None:
                self.aclv4_in_cfg_version = random.randint(
                    -9007199254740991, 9007199254740991
                )
        if list_type == "mac":
            self.aclmac_in_cfg = acl_obj
            if self.aclmac_in_cfg_version is None:
                self.aclmac_in_cfg_version = random.randint(
                    -9007199254740991, 9007199254740991
                )

        # Apply changes
        return self.apply()

    def attach_acl_out(self, acl_name, list_type):
        """
        Update ACL OUT values inside a Vlan object.
        :param acl_name: Alphanumeric String that is the name of the ACL.
        :param list_type: Alphanumeric String of ipv4, ipv6, or mac to specify
            the type of ACL.
        :return: True if object was changed.
        """
        # Create Acl object
        acl_obj = self.session.api.get_module(
            self.session, "ACL", index_id=acl_name, list_type=list_type
        )

        if list_type == "ipv6":
            self.aclv6_out_cfg = acl_obj
            if self.aclv6_out_cfg_version is None:
                self.aclv6_out_cfg_version = random.randint(
                    -9007199254740991, 9007199254740991
                )
        if list_type == "ipv4":
            self.aclv4_out_cfg = acl_obj
            if self.aclv4_out_cfg_version is None:
                self.aclv4_out_cfg_version = random.randint(
                    -9007199254740991, 9007199254740991
                )
        if list_type == "mac":
            self.aclmac_out_cfg = acl_obj
            if self.aclmac_out_cfg_version is None:
                self.aclmac_out_cfg_version = random.randint(
                    -9007199254740991, 9007199254740991
                )

        # Apply changes
        return self.apply()

    def detach_acl_in(self, acl_name, list_type):
        """
        Detach an ACL from a VLAN.
        :param acl_name: Alphanumeric String that is the name of the ACL.
        :param list_type: Alphanumeric String of ipv4, ipv6, or mac to specify
            the type of ACL.
        :return: True if object was changed.
        """
        if list_type == "ipv6":
            self.aclv6_in_cfg = None
            self.aclv6_in_cfg_version = None
        elif list_type == "ipv4":
            self.aclv4_in_cfg = None
            self.aclv4_in_cfg_version = None
        elif list_type == "mac":
            self.aclmac_in_cfg = None
            self.aclmac_in_cfg_version = None

        # Apply changes
        return self.apply()

    def detach_acl_out(self, acl_name, list_type):
        """
        Detach an ACL from a VLAN.
        :param acl_name: Alphanumeric String that is the name of the ACL.
        :param list_type: Alphanumeric String of ipv4, ipv6, or mac to specify
            the type of ACL.
        :return: True if object was changed.
        """
        if list_type == "ipv6":
            self.aclv6_out_cfg = None
            self.aclv6_out_cfg_version = None
        elif list_type == "ipv4":
            self.aclv4_out_cfg = None
            self.aclv4_out_cfg_version = None
        elif list_type == "mac":
            self.aclmac_out_cfg = None
            self.aclmac_out_cfg_version = None

        # Apply changes
        return self.apply()

    def get_mac(self, from_id, mac_address):
        """
        Create an Mac object.
        :param from_id: String source of the MAC address. Must be "dynamic",
            "VSX", "static", "VRRP", "port-access-security", "evpn", or "hsc".
        :param mac_address: String MAC address. Example: '01:02:03:04:05:06'
        :return: Mac object.
        """
        mac_obj = self.session.api.get_module(
            self.session,
            "Mac",
            from_id,
            mac_addr=mac_address,
            parent_vlan=self,
        )

        # Get MAC data
        mac_obj.get()
        return mac_obj

    def add_static_mac(self, port, mac_address):
        """
        Create an StaticMac object.
        :param port: String for the Port's name. Example: 1/1/1
        :param mac_address: String MAC address. Example: '01:02:03:04:05:06'
        :return: StaticMac object
        """
        if isinstance(port, str):
            # Make Interface into an object
            port = self.session.api.get_module(self.session, "Interface", port)
            # Materialize interface to ensure its existence
            port.get()

        static_mac_obj = self.session.api.get_module(
            self.session, "StaticMac", mac_address, parent_vlan=self, port=port
        )

        # Create static Mac on the switch
        static_mac_obj.apply()

        return static_mac_obj
